def print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times):
    print("Gantt Chart:")
    print("P\tAT\tBT\tWT\tTAT\tCT\tRT")
    for i in range(len(processes)):
        print(f"{processes[i]}\t{arrival_times[i]}\t{burst_times[i]}\t{wait_times[i]}\t{turn_around_times[i]}\t{completion_times[i]}\t{response_times[i]}")

    print(f"\nAverage Waiting Time: {sum(wait_times) / len(wait_times):.2f}")
    print(f"Average Turnaround Time: {sum(turn_around_times) / len(turn_around_times):.2f}")

def priority_non_preemptive(processes, arrival_times, burst_times, priorities):
    n = len(processes)
    completion_times = [0] * n
    wait_times = [0] * n
    turn_around_times = [0] * n
    response_times = [0] * n
    is_completed = [False] * n

    current_time = 0
    completed = 0

    while completed < n:
        max_priority = -1
        selected = -1
        for i in range(n):
            if arrival_times[i] <= current_time and not is_completed[i]:
                if priorities[i] > max_priority:
                    max_priority = priorities[i]
                    selected = i

        if selected == -1:
            current_time += 1
            continue

        current_time += burst_times[selected]
        completion_times[selected] = current_time
        turn_around_times[selected] = completion_times[selected] - arrival_times[selected]
        wait_times[selected] = turn_around_times[selected] - burst_times[selected]
        response_times[selected] = wait_times[selected]

        is_completed[selected] = True
        completed += 1

    print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times)

processes = ["A", "B", "C", "D"]
arrival_times = [0, 1, 2, 3]
burst_times = [5, 3, 8, 6]
priorities = [3, 2, 1, 4]
time_quantum = 2

print("\n--- Non-Preemptive Priority ---")
priority_non_preemptive(processes, arrival_times, burst_times, priorities)


def print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times):
    print("Gantt Chart:")
    print("P\tAT\tBT\tWT\tTAT\tCT\tRT")
    for i in range(len(processes)):
        print(f"{processes[i]}\t{arrival_times[i]}\t{burst_times[i]}\t{wait_times[i]}\t{turn_around_times[i]}\t{completion_times[i]}\t{response_times[i]}")

    print(f"\nAverage Waiting Time: {sum(wait_times) / len(wait_times):.2f}")
    print(f"Average Turnaround Time: {sum(turn_around_times) / len(turn_around_times):.2f}")

def sjf(processes, arrival_times, burst_times):
    n = len(processes)
    remaining_times = burst_times[:]
    completion_times = [0] * n
    wait_times = [0] * n
    turn_around_times = [0] * n
    response_times = [0] * n

    completed = 0
    current_time = 0
    is_completed = [False] * n

    while completed < n:
        shortest = -1
        min_burst = float('inf')
        for i in range(n):
            if arrival_times[i] <= current_time and not is_completed[i] and burst_times[i] < min_burst:
                shortest = i
                min_burst = burst_times[i]

        if shortest == -1:
            current_time += 1
            continue

        current_time += burst_times[shortest]
        completion_times[shortest] = current_time
        turn_around_times[shortest] = completion_times[shortest] - arrival_times[shortest]
        wait_times[shortest] = turn_around_times[shortest] - burst_times[shortest]
        response_times[shortest] = wait_times[shortest]

        is_completed[shortest] = True
        completed += 1

    print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times)

processes = ["A", "B", "C", "D"]
arrival_times = [0, 1, 2, 3]
burst_times = [5, 3, 8, 6]

print("\n--- SJF ---")
sjf(processes, arrival_times, burst_times)


def sstf(requests, head_start, total_tracks):
    sequence = []
    total_movement = 0
    current_head = head_start
    remaining = requests.copy()
    
    print("\nSSTF Disk Scheduling")
    print("-------------------")
    print("Sequence:", end=" ")
    
    while remaining:
        closest = min(remaining, key=lambda x: abs(x - current_head))
        movement = abs(closest - current_head)
        total_movement += movement
        current_head = closest
        remaining.remove(closest)
        sequence.append(closest)
        print(closest, end=" ")
    
    print("\nTotal head movement:", total_movement)
    print("Average head movement:", round(total_movement/len(requests), 2))
    return sequence, total_movement

requests = [45, 21, 67, 90, 4, 89, 52, 61, 87, 25]
head_start = 50
total_tracks = 100

print("Input Values:")
print("Request sequence:", requests)
print("Initial head position:", head_start)
print("Total tracks:", total_tracks)

sstf(requests.copy(), head_start, total_tracks)


def print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times):
    print("Gantt Chart:")
    print("P\tAT\tBT\tWT\tTAT\tCT\tRT")
    for i in range(len(processes)):
        print(f"{processes[i]}\t{arrival_times[i]}\t{burst_times[i]}\t{wait_times[i]}\t{turn_around_times[i]}\t{completion_times[i]}\t{response_times[i]}")

    print(f"\nAverage Waiting Time: {sum(wait_times) / len(wait_times):.2f}")
    print(f"Average Turnaround Time: {sum(turn_around_times) / len(turn_around_times):.2f}")
    
def preemptive_priority_round_robin(processes, arrival_times, burst_times, time_quantum, priorities):
    n = len(processes)
    remaining_times = burst_times[:]
    completion_times = [0] * n
    wait_times = [0] * n
    turn_around_times = [0] * n
    response_times = [-1] * n

    current_time = 0
    queue = []
    visited = [False] * n
    completed = 0

    # Combine processes based on priority
    while completed < n:
        # Find the process with the highest priority (lowest priority number) and not completed
        highest_priority_process = None
        for i in range(n):
            if arrival_times[i] <= current_time and remaining_times[i] > 0:
                if highest_priority_process is None or priorities[i] < priorities[highest_priority_process]:
                    highest_priority_process = i

        if highest_priority_process is None:
            current_time += 1
            continue

        # If a process is found, check if it's the first time it's being processed
        process_index = highest_priority_process
        if response_times[process_index] == -1:
            response_times[process_index] = current_time - arrival_times[process_index]

        # Execute the process using Round-Robin if it shares the same priority
        if remaining_times[process_index] > time_quantum:
            current_time += time_quantum
            remaining_times[process_index] -= time_quantum
        else:
            current_time += remaining_times[process_index]
            completion_times[process_index] = current_time
            remaining_times[process_index] = 0
            completed += 1

    for i in range(n):
        turn_around_times[i] = completion_times[i] - arrival_times[i]
        wait_times[i] = turn_around_times[i] - burst_times[i]

    print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times)

processes = ["A", "B", "C", "D"]
arrival_times = [0, 1, 2, 3]
burst_times = [5, 3, 8, 6]
time_quantum = 2
priorities = [1, 1, 2, 2]

print("\n--- Preemptive Round Robin ---")
preemptive_priority_round_robin(processes, arrival_times, burst_times, time_quantum, priorities)



from collections import deque

def fifo_replacement(pages, capacity):
    page_faults = 0
    frames = deque(maxlen=capacity)
    frame_states = []
    fault_list = []
    
    for page in pages:
        is_fault = page not in frames
        if is_fault:
            page_faults += 1
            if len(frames) == capacity:
                frames.popleft()
            frames.append(page)
        fault_list.append('F' if is_fault else 'H')
        frame_states.append(list(frames))
            
    return page_faults, frame_states, fault_list

def lru_replacement(pages, capacity):
    page_faults = 0
    frames = []
    page_usage = {}
    frame_states = []
    fault_list = []
    
    for time, page in enumerate(pages):
        is_fault = page not in frames
        if is_fault:
            page_faults += 1
            if len(frames) == capacity:
                lru_page = min(frames, key=lambda x: page_usage[x])
                frames.remove(lru_page)
            frames.append(page)
        
        page_usage[page] = time
        fault_list.append('F' if is_fault else 'H')
        frame_states.append(list(frames))
            
    return page_faults, frame_states, fault_list

def optimal_replacement(pages, capacity):
    page_faults = 0
    frames = []
    frame_states = []
    fault_list = []
    
    for i, page in enumerate(pages):
        is_fault = page not in frames
        if is_fault:
            page_faults += 1
            if len(frames) == capacity:
                future_usage = {}
                for frame in frames:
                    try:
                        future_usage[frame] = pages[i+1:].index(frame)
                    except ValueError:
                        future_usage[frame] = float('inf')
                
                victim_page = max(frames, key=lambda x: future_usage[x])
                frames.remove(victim_page)
            frames.append(page)
        fault_list.append('F' if is_fault else 'H')
        frame_states.append(list(frames))
            
    return page_faults, frame_states, fault_list

def print_page_table(algorithm_name, pages, frame_states, fault_list):
    print(f"\n{algorithm_name} Page Table:\n")
    
    print("Ref    :", end=" ")
    for page in pages:
        print(f"{page:2}", end=" ")
    print()
        
    print("" + "-" * (len(pages) * 3 + 10))
    
    for frame_idx in range(len(frame_states[-1])):
        print(f"Frame {frame_idx}:", end=" ")
        for state in frame_states:
            if frame_idx < len(state):
                print(f"{state[frame_idx]:2}", end=" ")
            else:
                print(" -", end=" ")
        print()

    print("" + "-" * (len(pages) * 3 + 10))

    print("Status :  ", end="")

    for status in fault_list:
        print(f"{status:2}", end=" ")


pages = [6, 1, 1, 2, 0, 3, 4, 6, 0, 2, 1, 2, 1, 2, 0, 3, 2, 1, 2, 0]
capacity = 3

fifo_faults, fifo_states, fifo_fault_list = fifo_replacement(pages, capacity)
lru_faults, lru_states, lru_fault_list = lru_replacement(pages, capacity)
optimal_faults, optimal_states, optimal_fault_list = optimal_replacement(pages, capacity)

print("Page Reference String:", pages)
print("Number of Frames:", capacity)

print_page_table("FIFO", pages, fifo_states, fifo_fault_list)
print("\n\nFIFO Page Faults:", fifo_faults, end="\n\n")

print_page_table("LRU", pages, lru_states, lru_fault_list)
print("\n\nLRU Page Faults:", lru_faults, end="\n\n")

print_page_table("Optimal", pages, optimal_states, optimal_fault_list)
print("\n\nOptimal Page Faults:", optimal_faults, end="\n\n")


def print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times):
    print("Gantt Chart:")
    print("P\tAT\tBT\tWT\tTAT\tCT\tRT")
    for i in range(len(processes)):
        print(f"{processes[i]}\t{arrival_times[i]}\t{burst_times[i]}\t{wait_times[i]}\t{turn_around_times[i]}\t{completion_times[i]}\t{response_times[i]}")

    print(f"\nAverage Waiting Time: {sum(wait_times) / len(wait_times):.2f}")
    print(f"Average Turnaround Time: {sum(turn_around_times) / len(turn_around_times):.2f}")

def srtf(processes, arrival_times, burst_times):
    n = len(processes)
    remaining_times = burst_times[:]
    completion_times = [0] * n
    wait_times = [0] * n
    turn_around_times = [0] * n
    response_times = [-1] * n

    current_time = 0
    completed = 0

    while completed < n:
        shortest = -1
        min_time = float('inf')
        for i in range(n):
            if arrival_times[i] <= current_time and remaining_times[i] > 0:
                if remaining_times[i] < min_time or (remaining_times[i] == min_time and arrival_times[i] < arrival_times[shortest]):
                    shortest = i
                    min_time = remaining_times[i]

        if shortest == -1:
            current_time += 1
            continue

        if response_times[shortest] == -1:
            response_times[shortest] = current_time - arrival_times[shortest]

        current_time += 1
        remaining_times[shortest] -= 1

        if remaining_times[shortest] == 0:
            completed += 1
            completion_times[shortest] = current_time

    for i in range(n):
        turn_around_times[i] = completion_times[i] - arrival_times[i]
        wait_times[i] = turn_around_times[i] - burst_times[i]

    print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times)


processes = ["A", "B", "C", "D"]
arrival_times = [0, 1, 2, 3]
burst_times = [5, 3, 8, 6]

print("\n--- SRTF ---")
srtf(processes, arrival_times, burst_times)


def print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times):
    print("Gantt Chart:")
    print("P\tAT\tBT\tWT\tTAT\tCT\tRT")
    for i in range(len(processes)):
        print(f"{processes[i]}\t{arrival_times[i]}\t{burst_times[i]}\t{wait_times[i]}\t{turn_around_times[i]}\t{completion_times[i]}\t{response_times[i]}")

    print(f"\nAverage Waiting Time: {sum(wait_times) / len(wait_times):.2f}")
    print(f"Average Turnaround Time: {sum(turn_around_times) / len(turn_around_times):.2f}")

def fcfs(processes, arrival_times, burst_times):
    n = len(processes)
    completion_times = [0] * n
    wait_times = [0] * n
    turn_around_times = [0] * n
    response_times = [0] * n

    completion_times[0] = arrival_times[0] + burst_times[0]
    for i in range(1, n):
        completion_times[i] = max(completion_times[i - 1], arrival_times[i]) + burst_times[i]

    for i in range(n):
        turn_around_times[i] = completion_times[i] - arrival_times[i]
        wait_times[i] = turn_around_times[i] - burst_times[i]
        response_times[i] = wait_times[i]

    print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times)

processes = ["A", "B", "C", "D"]
arrival_times = [0, 1, 2, 3]
burst_times = [5, 3, 8, 6]
priorities = [3, 2, 1, 4]
time_quantum = 2

print("\n--- FCFS ---")
fcfs(processes, arrival_times, burst_times)


# os-sem1/module.py

def scan(initial, requests, disk_size, direction):
    requests.sort()
    left = [r for r in requests if r < initial]
    right = [r for r in requests if r >= initial]

    if direction == "up":
        sequence = right + [disk_size - 1] + left[::-1]
    else:
        sequence = left[::-1] + [0] + right

    seek_time = 0
    current = initial
    for pos in sequence:
        seek_time += abs(pos - current)
        current = pos

    print(f"SCAN Order: {sequence}")
    print(f"Total Seek Time: {seek_time}")


requests = [82, 170, 43, 140, 24, 16, 190]
initial = 50
disk_size = 200
direction = "up"

scan(initial, requests, disk_size, direction)










def fifo(pages, frames):
    memory = []
    page_faults = 0

    for page in pages:
        if page not in memory:
            if len(memory) >= frames:
                memory.pop(0)
            memory.append(page)
            page_faults += 1

    return page_faults

def lru(pages, frames):
    memory = []
    page_faults = 0

    for page in pages:
        if page not in memory:
            if len(memory) >= frames:
                memory.pop(-1)
            page_faults += 1
    return

def optimal(pages, frames):
    memory = []
    page_faults = 0

    for i, page in enumerate(pages):
        if page not in memory:
            if len(memory) < frames:
                memory.append(page)
            else:
                farthest = -1
                victim = -1
                for mem_page in memory:
                    if mem_page not in pages[i+1:]:
                        victim = mem_page
                        break
                    else:
                        idx = pages[i+1:].index(mem_page)
                        if idx > farthest:
                            farthest = idx
                            victim = mem_page
                memory[memory.index(victim)] = page
            page_faults += 1

    return page_faults

# pages = [7, 0, 1, 2, 0, 3, 0, 4, 2, 3, 0, 3, 2]
# frames = 3

# print("FIFO Page Faults:", fifo(pages, frames))
# print("LRU Page Faults:", lru(pages, frames))
# print("Optimal Page Faults:", optimal(pages, frames))


def print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times):
    print("Gantt Chart:")
    print("P\tAT\tBT\tWT\tTAT\tCT\tRT")
    for i in range(len(processes)):
        print(f"{processes[i]}\t{arrival_times[i]}\t{burst_times[i]}\t{wait_times[i]}\t{turn_around_times[i]}\t{completion_times[i]}\t{response_times[i]}")

    print(f"\nAverage Waiting Time: {sum(wait_times) / len(wait_times):.2f}")
    print(f"Average Turnaround Time: {sum(turn_around_times) / len(turn_around_times):.2f}")

def round_robin(processes, arrival_times, burst_times, time_quantum):
    n = len(processes)
    remaining_times = burst_times[:]
    completion_times = [0] * n
    wait_times = [0] * n
    turn_around_times = [0] * n
    response_times = [-1] * n

    current_time = 0
    queue = []
    visited = [False] * n

    while True:
        all_completed = True
        for i in range(n):
            if arrival_times[i] <= current_time and not visited[i]:
                queue.append(i)
                visited[i] = True

        if queue:
            all_completed = False
            process_index = queue.pop(0)

            if response_times[process_index] == -1:
                response_times[process_index] = current_time - arrival_times[process_index]

            if remaining_times[process_index] > time_quantum:
                current_time += time_quantum
                remaining_times[process_index] -= time_quantum
                queue.append(process_index)
            else:
                current_time += remaining_times[process_index]
                completion_times[process_index] = current_time
                remaining_times[process_index] = 0
        elif all_completed:
            break
        else:
            current_time += 1

    for i in range(n):
        turn_around_times[i] = completion_times[i] - arrival_times[i]
        wait_times[i] = turn_around_times[i] - burst_times[i]

    print_gantt_chart(processes, arrival_times, burst_times, wait_times, turn_around_times, completion_times, response_times)

processes = ["A", "B", "C", "D"]
arrival_times = [0, 1, 2, 3]
burst_times = [5, 3, 8, 6]
time_quantum = 2

print("\n--- Round Robin ---")
round_robin(processes, arrival_times, burst_times, time_quantum)


# os-sem1/__init__.py

from .pagereplacement import fifo_replacement, lru_replacement, optimal_replacement

from .diskfcfs import diskfcfs
from .disksstf import disksstf
from .diskscan import diskscan
from .diskcscan import diskcscan

from .cpufcfs import cpufcfs
from .cpusrtf import cpusrtf
from .cpusjf import cpusjf
from .cpuroundrobin import cpuroundrobin
from .cpunonpreemptive import cpunonpreemptive
from .cpupreemptive import cpupreemptive


def scan(requests, head_start, total_tracks, direction="right"):
    sequence = []
    total_movement = 0
    current_head = head_start
    remaining = sorted(requests)
    
    print("\nSCAN Disk Scheduling")
    print("-------------------")
    print("Direction:", direction)
    print("Sequence:", end=" ")
    
    if direction == "right":
        for request in [r for r in remaining if r >= current_head]:
            movement = abs(request - current_head)
            total_movement += movement
            current_head = request
            sequence.append(request)
            print(request, end=" ")
        
        if current_head < total_tracks - 1:
            total_movement += (total_tracks - 1 - current_head)
            current_head = total_tracks - 1
            print(total_tracks - 1, end=" ")
        
        for request in [r for r in remaining if r < head_start][::-1]:
            movement = abs(request - current_head)
            total_movement += movement
            current_head = request
            sequence.append(request)
            print(request, end=" ")
    
    print("\nTotal head movement:", total_movement)
    print("Average head movement:", round(total_movement/len(requests), 2))
    return sequence, total_movement

requests = [45, 21, 67, 90, 4, 89, 52, 61, 87, 25]
head_start = 50
total_tracks = 100

print("Input Values:")
print("Request sequence:", requests)
print("Initial head position:", head_start)
print("Total tracks:", total_tracks)

scan(requests.copy(), head_start, total_tracks, "right")